/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * fische-3.1
 * Copyright (C) Marcel Ebmer 2009-2010 <marcel@26elf.at>
 *
 * fische-3.1 is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * fische-3.1 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef WINDOWS_BUILD
// please note that fische will not compile or run on windows
// this is for development reasons only
#include <SDL_main.h>
#endif // WINDOWS_BUILD

#include <SDL/SDL.h>
#include <iostream>
#include <cstdlib>
#include <unistd.h>
#include <pthread.h>
#include <stdint.h>
#include "sdlscreen.h"
#include "vectorfield.h"
#include "ringbuffer.h"
#include "blurengine.h"
#include "alsa-samples.h"
#include "pulseaudio-samples.h"
#include "portaudio-samples.h"
#include "dummyaudio-samples.h"
#include "wavepainter.h"
#include "analyst.h"
#include "configuration.h"

using namespace std;

int main (int argc, char** argv) {

	bool paused = false;
	uint32_t frames = 0;
	uint32_t ticks = 0;

	// initialize random number generator
	srand (time (NULL) );

	// read options from commandline (or file - TODO)
	Configuration* configuration = Configuration::instance (argc, argv);

	// create ringbuffer
	RingBuffer* ringbuffer = RingBuffer::instance (173460 / configuration->fpsTarget() );

	// create sampling thread
	static bool run = true;

#ifndef WINDOWS_BUILD
	pthread_t samplingThread;
#else
	SDL_Thread* samplingThread;
#endif

	sample_param_t p;
	p.device = configuration->device();
	p.ringbuffer = ringbuffer;
	p.run = &run;

#ifndef WINDOWS_BUILD
	switch (configuration->audioDriver() ) {
		case Configuration::ALSA:
			pthread_create (&samplingThread, NULL, sample_alsa_pcm, (void*) &p);
			break;
		case Configuration::PULSE:
			pthread_create (&samplingThread, NULL, sample_pulseaudio, (void*) &p);
			break;
		case Configuration::PORTAUDIO:
			pthread_create (&samplingThread, NULL, sample_portaudio, (void*) &p);
			break;
		case Configuration::DUMMY:
		default:
			pthread_create (&samplingThread, NULL, sample_dummyaudio, (void*) &p);
			break;
	}
#else
	switch (configuration->audioDriver() ) {
		case Configuration::PORTAUDIO:
			samplingThread = SDL_CreateThread (sample_portaudio_win32, &p);
			break;
		case Configuration::DUMMY:
		default:
			samplingThread = SDL_CreateThread (sample_dummyaudio_win32, &p);
			break;
	}
#endif // WINDOWS_BUILD

	// create components
	VectorField* vectorfield = VectorField::instance (configuration->X(), configuration->Y() );
	SdlScreen* sdlscreen = SdlScreen::instance (configuration->screenX(), configuration->screenY(), configuration->fullscreen() );
	if ( (vectorfield->width() > sdlscreen->width() ) || (vectorfield->height() > sdlscreen->height() ) ) {
		cerr << "ERROR: animation is bigger than screen can allow." << endl;
		exit (EXIT_FAILURE);
	}
	BlurEngine* blurengine = BlurEngine::instance();
	WavePainter* wavepainter = WavePainter::instance();
	Analyst* analyst = Analyst::instance();

	int events;
	int analysis = 0;
	// main loop
	while (! ( (events = sdlscreen->checkEvents() ) & EXIT) ) {
		// check for mouse exit event if enabled
		if ( (events & MOUSE) && configuration->exitOnMouse() ) break;

		// check for interesting events
		if (events & TOGGLE_FULLSCREEN) {
			sdlscreen->toggleFullscreen();
			configuration->toggleFullscreen();
		}
		if (events & PAUSE) paused = paused ? false : true;

		// paused? do nothing.
		if (paused) {
			usleep (10000);
			continue;
		}

		// wait until ringbuffer is full
		int sleepcount = 0;
		while (!ringbuffer->ready() ) {
			usleep (1000);
			sleepcount ++;
			if (sleepcount > 100) break;
		}

		// ringbuffer not filling? in the meantime, check for new events.
		if (sleepcount > 100) continue;

		// analyse sound data
		if (configuration->audioDriver() != Configuration::DUMMY) {
			analysis = analyst->analyse();
			if (analysis >= 1) wavepainter->changeShape();
			if (analysis >= 2) vectorfield->change();
			if (analysis >= 3) wavepainter->beat (analyst->bpm() );
			if (analysis >= 4) wavepainter->changeColor();
		} else {
			if (frames % 15 == 0) wavepainter->changeShape();
			if (frames % 150 == 0) vectorfield->change();
			if (frames % 450 == 0) wavepainter->beat (15);
			if (frames % 1500 == 0) wavepainter->changeColor();
		}

		// have fun
		if (analysis != -1)
			wavepainter->paint();
		sdlscreen->update();
		blurengine->blur();
		frames ++;
		if (frames == 1) ticks = SDL_GetTicks();
	}

	// print out actual frames per second
	cout << "\n* actual fps was " << (double) (frames - 1) / (double) (SDL_GetTicks() - ticks) * 1000 << endl;

	// tell the sound sampling thread to quit
	run = false;
#ifndef WINDOWS_BUILD
	pthread_join (samplingThread, NULL);
#else
	SDL_WaitThread (samplingThread, NULL);
#endif

	analyst->release();
	blurengine->release();
	sdlscreen->release();
	vectorfield->release();
	ringbuffer->release();
	configuration->release();

	exit (EXIT_SUCCESS);
}
